from __future__ import annotations

import itertools
from contextlib import suppress
from typing import TYPE_CHECKING

from hypothesis import given
from hypothesis.strategies import booleans, sampled_from
from pytest import raises

from utilities.atomicwrites import (
    _MoveDirectoryExistsError,
    _MoveFileExistsError,
    _MoveSourceNotFoundError,
    _WriterDirectoryExistsError,
    _WriterFileExistsError,
    _WriterTemporaryPathEmptyError,
    move,
    move_many,
    writer,
)
from utilities.hypothesis import settings_with_reduced_examples, temp_paths

if TYPE_CHECKING:
    from pathlib import Path

    from utilities.types import OpenMode


class TestMove:
    @given(root=temp_paths(), overwrite=booleans())
    def test_file_destination_does_not_exist(
        self, *, root: Path, overwrite: bool
    ) -> None:
        source = root.joinpath("source")
        with source.open(mode="w") as fh:
            _ = fh.write("text")
        destination = root.joinpath("destination")
        move(source, destination, overwrite=overwrite)
        assert destination.is_file()
        with destination.open() as fh:
            assert fh.read() == "text"

    @given(root=temp_paths())
    def test_file_destination_file_exists(self, *, root: Path) -> None:
        source = root.joinpath("source")
        with source.open(mode="w") as fh:
            _ = fh.write("source")
        destination = root.joinpath("destination")
        with destination.open(mode="w") as fh:
            _ = fh.write("destination")
        move(source, destination, overwrite=True)
        assert destination.is_file()
        with destination.open() as fh:
            assert fh.read() == "source"

    @given(root=temp_paths())
    def test_file_destination_directory_exists(self, *, root: Path) -> None:
        source = root.joinpath("source")
        with source.open(mode="w") as fh:
            _ = fh.write("source")
        destination = root.joinpath("destination")
        destination.mkdir()
        move(source, destination, overwrite=True)
        assert destination.is_file()
        with destination.open() as fh:
            assert fh.read() == "source"

    @given(root=temp_paths(), overwrite=booleans())
    def test_directory_destination_does_not_exist(
        self, *, root: Path, overwrite: bool
    ) -> None:
        source = root.joinpath("source")
        source.mkdir()
        source.joinpath("file").touch()
        destination = root.joinpath("destination")
        move(source, destination, overwrite=overwrite)
        assert destination.is_dir()
        assert len(list(destination.iterdir())) == 1

    @given(root=temp_paths())
    def test_directory_destination_file_exists(self, *, root: Path) -> None:
        source = root.joinpath("source")
        source.mkdir()
        source.joinpath("file").touch()
        destination = root.joinpath("destination")
        destination.touch()
        move(source, destination, overwrite=True)
        assert destination.is_dir()
        assert len(list(destination.iterdir())) == 1

    @given(root=temp_paths())
    def test_directory_destination_directory_exists(self, *, root: Path) -> None:
        source = root.joinpath("source")
        source.mkdir()
        source.joinpath("file").touch()
        destination = root.joinpath("destination")
        destination.mkdir()
        for i in range(2):
            destination.joinpath(f"file{i}").touch()
        move(source, destination, overwrite=True)
        assert destination.is_dir()
        assert len(list(destination.iterdir())) == 1

    @given(root=temp_paths(), overwrite=booleans())
    def test_error_source_not_found(self, *, root: Path, overwrite: bool) -> None:
        with raises(_MoveSourceNotFoundError, match="Source '.*' does not exist"):
            move(
                root.joinpath("source"),
                root.joinpath("destination"),
                overwrite=overwrite,
            )

    @given(root=temp_paths())
    def test_error_file_exists(self, *, root: Path) -> None:
        source = root.joinpath("source")
        source.touch()
        destination = root.joinpath("destination")
        destination.touch()
        with raises(
            _MoveFileExistsError,
            match="Cannot move file '.*' as destination '.*' already exists",
        ):
            move(source, destination)

    @given(root=temp_paths())
    def test_error_directory_exists(self, *, root: Path) -> None:
        source = root.joinpath("source")
        source.mkdir()
        destination = root.joinpath("destination")
        destination.touch()
        with raises(
            _MoveDirectoryExistsError,
            match="Cannot move directory '.*' as destination '.*' already exists",
        ):
            move(source, destination)


class TestMoveMany:
    @given(root=temp_paths())
    @settings_with_reduced_examples()
    def test_many(self, *, root: Path) -> None:
        n = 5
        files = [root.joinpath(f"file{i}") for i in range(n + 1)]
        for i in range(n):
            with files[i].open(mode="w") as fh:
                _ = fh.write(str(i))
        move_many(*itertools.pairwise(files), overwrite=True)
        for i in range(1, n + 1):
            with files[i].open() as fh:
                assert fh.read() == str(i - 1)


class TestWriter:
    @given(
        root=temp_paths(),
        case=sampled_from([("w", "r", "contents"), ("wb", "rb", b"contents")]),
    )
    def test_file_writing(
        self, *, root: Path, case: tuple[OpenMode, OpenMode, str | bytes]
    ) -> None:
        write_mode, read_mode, contents = case
        path = root.joinpath("file.txt")
        with writer(path) as temp, temp.open(mode=write_mode) as fh1:
            _ = fh1.write(contents)
        with path.open(mode=read_mode) as fh2:
            assert fh2.read() == contents

    @given(root=temp_paths())
    def test_error_temporary_path_empty(self, *, root: Path) -> None:
        with (
            raises(
                _WriterTemporaryPathEmptyError, match="Temporary path '.*' is empty"
            ),
            writer(root),
        ):
            pass

    @given(root=temp_paths())
    def test_error_file_exists(self, *, root: Path) -> None:
        path = root.joinpath("file.txt")
        path.touch()
        with (
            raises(
                _WriterFileExistsError,
                match="Cannot write to '.*' as file already exists",
            ),
            writer(path) as temp,
            temp.open(mode="w") as fh,
        ):
            _ = fh.write("new contents")

    @given(root=temp_paths())
    def test_error_directory_exists(self, *, root: Path) -> None:
        path = root.joinpath("dir")
        path.mkdir()
        with (
            raises(
                _WriterDirectoryExistsError,
                match="Cannot write to '.*' as directory already exists",
            ),
            writer(path) as temp,
        ):
            temp.mkdir()

    @given(
        root=temp_paths(),
        case=sampled_from([(KeyboardInterrupt, False), (ValueError, True)]),
    )
    def test_error_during_write(
        self, *, root: Path, case: tuple[type[Exception], bool]
    ) -> None:
        error, expected = case
        path = root.joinpath("file.txt")

        def raise_error() -> None:
            raise error

        with writer(path) as temp1, temp1.open(mode="w") as fh, suppress(Exception):
            _ = fh.write("contents")
            raise_error()
        is_non_empty = len(list(root.iterdir())) >= 1
        assert is_non_empty is expected



from __future__ import annotations

from typing import TYPE_CHECKING

from polars import struct
from polars_ols import RollingKwargs, compute_rolling_least_squares

from utilities.polars import ensure_expr_or_series

if TYPE_CHECKING:
    from polars import Expr
    from polars_ols import NullPolicy

    from utilities.polars import ExprLike


def compute_rolling_ols(
    target: ExprLike,
    *features: ExprLike,
    sample_weights: ExprLike | None = None,
    add_intercept: bool = False,
    null_policy: NullPolicy = "drop_window",
    window_size: int = 1000000,
    min_periods: int | None = None,
    use_woodbury: bool | None = None,
    alpha: float | None = None,
) -> Expr:
    """Compute a rolling OLS."""
    target = ensure_expr_or_series(target)
    rolling_kwargs = RollingKwargs(
        null_policy=null_policy,
        window_size=window_size,
        min_periods=min_periods,
        use_woodbury=use_woodbury,
        alpha=alpha,
    )
    coefficients = compute_rolling_least_squares(
        target,
        *features,
        sample_weights=sample_weights,
        add_intercept=add_intercept,
        mode="coefficients",
        rolling_kwargs=rolling_kwargs,
    ).alias("coefficients")
    predictions = compute_rolling_least_squares(
        target,
        *features,
        sample_weights=sample_weights,
        add_intercept=add_intercept,
        mode="predictions",
        rolling_kwargs=rolling_kwargs,
    ).alias("predictions")
    residuals = compute_rolling_least_squares(
        target,
        *features,
        sample_weights=sample_weights,
        add_intercept=add_intercept,
        mode="residuals",
        rolling_kwargs=rolling_kwargs,
    ).alias("residuals")
    ssr = (residuals**2).rolling_sum(window_size, min_samples=min_periods).alias("SSR")
    sst = (
        ((target - target.rolling_mean(window_size, min_samples=min_periods)) ** 2)
        .rolling_sum(window_size, min_samples=min_periods)
        .alias("SST")
    )
    r2 = (1 - ssr / sst).alias("R2")
    return struct(coefficients, predictions, residuals, r2).alias("ols")


__all__ = ["compute_rolling_ols"]



from __future__ import annotations

import datetime as dt
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path
from types import NoneType
from typing import Literal

from hypothesis import given
from hypothesis.strategies import booleans, dates, floats, integers, sampled_from, times
from pytest import raises

from tests.test_operator import TruthEnum
from tests.test_typing_funcs.with_future import (
    DataClassFutureInt,
    DataClassFutureIntEven,
    DataClassFutureIntEvenOrOddTypeUnion,
    DataClassFutureIntEvenOrOddUnion,
    DataClassFutureIntOdd,
    TrueOrFalseFutureLit,
    TrueOrFalseFutureTypeLit,
)
from utilities.errors import ImpossibleCaseError
from utilities.functions import ensure_path
from utilities.hypothesis import (
    datetime_durations,
    local_datetimes,
    numbers,
    paths,
    text_ascii,
    timedeltas_2w,
    versions,
    zoned_datetimes,
)
from utilities.math import is_equal
from utilities.parse import (
    _ParseTextExtraNonUniqueError,
    _ParseTextParseError,
    parse_text,
)
from utilities.sentinel import Sentinel, sentinel
from utilities.types import Duration, Number
from utilities.version import Version
from utilities.whenever import (
    serialize_date,
    serialize_datetime,
    serialize_duration,
    serialize_time,
    serialize_timedelta,
)


class TestParseText:
    @given(value=booleans())
    def test_bool(self, *, value: bool) -> None:
        text = str(value)
        result = parse_text(bool, text)
        assert result is value

    @given(date=dates())
    def test_date(self, *, date: dt.date) -> None:
        text = serialize_date(date)
        result = parse_text(dt.date, text)
        assert result == date

    @given(datetime=local_datetimes() | zoned_datetimes())
    def test_datetime(self, *, datetime: dt.datetime) -> None:
        text = serialize_datetime(datetime)
        result = parse_text(dt.datetime, text)
        assert result == datetime

    @given(duration=datetime_durations(two_way=True))
    def test_duration(self, *, duration: Duration) -> None:
        text = serialize_duration(duration)
        result = parse_text(Duration, text)
        assert result == duration

    @given(truth=sampled_from(TruthEnum))
    def test_enum(self, *, truth: TruthEnum) -> None:
        text = truth.name
        result = parse_text(TruthEnum, text)
        assert result is truth

    @given(value=integers())
    def test_extra_type(self, *, value: int) -> None:
        text = str(value)
        result = parse_text(
            DataClassFutureInt,
            text,
            extra={DataClassFutureInt: lambda text: DataClassFutureInt(int_=int(text))},
        )
        expected = DataClassFutureInt(int_=value)
        assert result == expected

    @given(value=floats())
    def test_float(self, *, value: float) -> None:
        text = str(value)
        result = parse_text(float, text)
        assert is_equal(result, value)

    @given(value=integers())
    def test_int(self, *, value: int) -> None:
        text = str(value)
        result = parse_text(int, text)
        assert result == value

    @given(truth=sampled_from(["true", "false"]))
    def test_literal(self, *, truth: Literal["true", "false"]) -> None:
        result = parse_text(TrueOrFalseFutureLit, truth)
        assert result == truth

    def test_none(self) -> None:
        text = str(None)
        result = parse_text(None, text)
        assert result is None

    def test_none_type(self) -> None:
        text = str(None)
        result = parse_text(NoneType, text)
        assert result is None

    @given(number=numbers())
    def test_number(self, *, number: Number) -> None:
        text = str(number)
        result = parse_text(Number, text)
        assert result == number

    @given(path=paths())
    def test_path(self, *, path: Path) -> None:
        text = str(path)
        result = parse_text(Path, text)
        assert result == path

    @given(path=paths())
    def test_path_expanded(self, *, path: Path) -> None:
        path_use = Path("~", path)
        text = str(path_use)
        result = ensure_path(parse_text(Path, text))
        assert result == result.expanduser()

    def test_nullable_number_none(self) -> None:
        text = str(None)
        result = parse_text(Number | None, text)
        assert result is None

    @given(number=numbers())
    def test_nullable_number_number(self, *, number: Number) -> None:
        text = str(number)
        result = parse_text(Number | None, text)
        assert result == number

    def test_nullable_duration_none(self) -> None:
        text = str(None)
        result = parse_text(Duration | None, text)
        assert result is None

    @given(duration=datetime_durations(two_way=True))
    def test_nullable_duration_duration(self, *, duration: Duration) -> None:
        text = serialize_duration(duration)
        result = parse_text(Duration | None, text)
        assert result == duration

    def test_nullable_int_none(self) -> None:
        text = str(None)
        result = parse_text(int | None, text)
        assert result is None

    @given(value=integers())
    def test_nullable_int_int(self, *, value: int) -> None:
        text = str(value)
        result = parse_text(int | None, text)
        assert result == value

    def test_sentinel(self) -> None:
        text = str(sentinel)
        result = parse_text(Sentinel, text)
        assert result is sentinel

    @given(text=text_ascii())
    def test_str(self, *, text: str) -> None:
        result = parse_text(str, text)
        assert result == text

    @given(time=times())
    def test_time(self, *, time: dt.time) -> None:
        text = serialize_time(time)
        result = parse_text(dt.time, text)
        assert result == time

    @given(timedelta=timedeltas_2w())
    def test_timedelta(self, *, timedelta: dt.timedelta) -> None:
        text = serialize_timedelta(timedelta)
        result = parse_text(dt.timedelta, text)
        assert result == timedelta

    @given(x=integers(), y=integers())
    def test_tuple(self, *, x: int, y: int) -> None:
        text = f"({x}, {y})"
        result = parse_text(tuple[int, int], text)
        assert result == (x, y)

    @given(truth=sampled_from(["true", "false"]))
    def test_type_literal(self, *, truth: Literal["true", "false"]) -> None:
        result = parse_text(TrueOrFalseFutureTypeLit, truth)
        assert result == truth

    @given(value=integers())
    def test_type_union_with_extra(self, *, value: int) -> None:
        def parse_even_or_odd(text: str, /) -> DataClassFutureIntEvenOrOddTypeUnion:
            value = int(text)
            match value % 2:
                case 0:
                    return DataClassFutureIntEven(even_int=value)
                case 1:
                    return DataClassFutureIntOdd(odd_int=value)
                case _:
                    raise ImpossibleCaseError(case=[f"{value=}"])

        text = str(value)
        result = parse_text(
            DataClassFutureIntEvenOrOddTypeUnion,
            text,
            extra={DataClassFutureIntEvenOrOddTypeUnion: parse_even_or_odd},
        )
        match value % 2:
            case 0:
                expected = DataClassFutureIntEven(even_int=value)
            case 1:
                expected = DataClassFutureIntOdd(odd_int=value)
            case _:
                raise ImpossibleCaseError(case=[f"{value=}"])
        assert result == expected

    @given(value=integers())
    def test_union_with_extra(self, *, value: int) -> None:
        def parse_even_or_odd(text: str, /) -> DataClassFutureIntEvenOrOddUnion:
            value = int(text)
            match value % 2:
                case 0:
                    return DataClassFutureIntEven(even_int=value)
                case 1:
                    return DataClassFutureIntOdd(odd_int=value)
                case _:
                    raise ImpossibleCaseError(case=[f"{value=}"])

        text = str(value)
        result = parse_text(
            DataClassFutureIntEvenOrOddUnion,
            text,
            extra={DataClassFutureIntEvenOrOddUnion: parse_even_or_odd},
        )
        match value % 2:
            case 0:
                expected = DataClassFutureIntEven(even_int=value)
            case 1:
                expected = DataClassFutureIntOdd(odd_int=value)
            case _:
                raise ImpossibleCaseError(case=[f"{value=}"])
        assert result == expected

    @given(version=versions())
    def test_version(self, *, version: Version) -> None:
        text = str(version)
        result = parse_text(Version, text)
        assert result == version

    def test_error_bool(self) -> None:
        with raises(
            _ParseTextParseError, match="Unable to parse <class 'bool'>; got 'invalid'"
        ):
            _ = parse_text(bool, "invalid")

    def test_error_date(self) -> None:
        with raises(
            _ParseTextParseError,
            match=r"Unable to parse <class 'datetime\.date'>; got 'invalid'",
        ):
            _ = parse_text(dt.date, "invalid")

    def test_error_datetime(self) -> None:
        with raises(
            _ParseTextParseError,
            match=r"Unable to parse <class 'datetime\.datetime'>; got 'invalid'",
        ):
            _ = parse_text(dt.datetime, "invalid")

    def test_error_duration(self) -> None:
        with raises(
            _ParseTextParseError, match=r"Unable to parse Duration; got 'invalid'"
        ):
            _ = parse_text(Duration, "invalid")

    def test_error_enum(self) -> None:
        with raises(
            _ParseTextParseError,
            match="Unable to parse <enum 'TruthEnum'>; got 'invalid'",
        ):
            _ = parse_text(TruthEnum, "invalid")

    def test_error_extra_empty(self) -> None:
        with raises(
            _ParseTextParseError,
            match="Unable to parse <class 'tests.test_typing_funcs.with_future.DataClassFutureInt'>; got 'invalid'",
        ):
            _ = parse_text(DataClassFutureInt, "invalid", extra={})

    @given(value=integers())
    def test_error_extra_non_unique(self, *, value: int) -> None:
        @dataclass(kw_only=True)
        class Parent1:
            x: int = 0

        @dataclass(kw_only=True)
        class Parent2:
            y: int = 0

        @dataclass(kw_only=True)
        class Child(Parent1, Parent2): ...

        with raises(
            _ParseTextExtraNonUniqueError,
            match="Unable to parse <class '.*'> since `extra` must contain exactly one parent class; got <function .*>, <function .*> and perhaps more",
        ):
            _ = parse_text(
                Child,
                str(value),
                extra={
                    Parent1: lambda text: Child(x=int(text)),
                    Parent2: lambda text: Child(y=int(text)),
                },
            )

    def test_error_union_type_extra(self) -> None:
        with raises(
            _ParseTextParseError,
            match=r"Unable to parse tests\.test_typing_funcs\.with_future\.DataClassFutureIntEven \| tests\.test_typing_funcs\.with_future\.DataClassFutureIntOdd; got 'invalid'",
        ):
            _ = parse_text(DataClassFutureIntEvenOrOddUnion, "invalid", extra={})

    def test_error_float(self) -> None:
        with raises(
            _ParseTextParseError, match="Unable to parse <class 'float'>; got 'invalid'"
        ):
            _ = parse_text(float, "invalid")

    def test_error_int(self) -> None:
        with raises(
            _ParseTextParseError, match="Unable to parse <class 'int'>; got 'invalid'"
        ):
            _ = parse_text(int, "invalid")

    def test_error_none(self) -> None:
        with raises(_ParseTextParseError, match="Unable to parse None; got 'invalid'"):
            _ = parse_text(None, "invalid")

    def test_error_none_type(self) -> None:
        with raises(
            _ParseTextParseError,
            match="Unable to parse <class 'NoneType'>; got 'invalid'",
        ):
            _ = parse_text(NoneType, "invalid")

    def test_error_nullable_int(self) -> None:
        with raises(
            _ParseTextParseError, match=r"Unable to parse int \| None; got 'invalid'"
        ):
            _ = parse_text(int | None, "invalid")

    def test_error_nullable_not_type(self) -> None:
        with raises(
            _ParseTextParseError,
            match=r"Unable to parse collections\.abc\.Iterable\[None\] \| None; got 'invalid'",
        ):
            _ = parse_text(Iterable[None] | None, "invalid")

    def test_error_number(self) -> None:
        with raises(
            _ParseTextParseError, match=r"Unable to parse Number; got 'invalid'"
        ):
            _ = parse_text(Number, "invalid")

    def test_error_sentinel(self) -> None:
        with raises(
            _ParseTextParseError,
            match=r"Unable to parse <class 'utilities\.sentinel\.Sentinel'>; got 'invalid'",
        ):
            _ = parse_text(Sentinel, "invalid")

    def test_error_time(self) -> None:
        with raises(
            _ParseTextParseError,
            match=r"Unable to parse <class 'datetime\.time'>; got 'invalid'",
        ):
            _ = parse_text(dt.time, "invalid")

    def test_error_timedelta(self) -> None:
        with raises(
            _ParseTextParseError,
            match=r"Unable to parse <class 'datetime\.timedelta'>; got 'invalid'",
        ):
            _ = parse_text(dt.timedelta, "invalid")

    def test_error_tuple_invalid_text(self) -> None:
        with raises(
            _ParseTextParseError,
            match=r"Unable to parse tuple\[int, int\]; got 'invalid'",
        ):
            _ = parse_text(tuple[int, int], "invalid")

    def test_error_tuple_inconsistent_args_and_texts(self) -> None:
        with raises(
            _ParseTextParseError,
            match=r"Unable to parse tuple\[int, int\]; got '\(text1, text2, text3\)'",
        ):
            _ = parse_text(tuple[int, int], "(text1, text2, text3)")

    def test_error_unknown_annotation(self) -> None:
        with raises(
            _ParseTextParseError, match=r"Unable to parse int \| str; got 'invalid'"
        ):
            _ = parse_text(int | str, "invalid")

    def test_error_unknown_type(self) -> None:
        with raises(
            _ParseTextParseError,
            match=r"Unable to parse <class 'tests\.test_typing_funcs\.with_future\.DataClassFutureInt'>; got 'invalid'",
        ):
            _ = parse_text(DataClassFutureInt, "invalid")

    def test_error_unknown_union_type(self) -> None:
        with raises(
            _ParseTextParseError,
            match=r"Unable to parse tests\.test_typing_funcs\.with_future\.DataClassFutureIntEven \| tests\.test_typing_funcs\.with_future\.DataClassFutureIntOdd; got 'invalid'",
        ):
            _ = parse_text(DataClassFutureIntEvenOrOddUnion, "invalid")

    def test_error_version(self) -> None:
        with raises(
            _ParseTextParseError,
            match=r"Unable to parse <class 'utilities\.version\.Version'>; got 'invalid'",
        ):
            _ = parse_text(Version, "invalid")



from __future__ import annotations

from itertools import chain

from utilities.traceback import trace


@trace
def func_many(a: int, b: int, /, *args: int, c: int = 0, **kwargs: int) -> int:
    a *= 2
    b *= 2
    args = tuple(2 * arg for arg in args) + tuple(2 * arg for arg in range(100))
    c *= 2
    kwargs = {k: 2 * v for k, v in kwargs.items()}
    result = sum(chain([a, b], args, [c], kwargs.values()))
    assert result % 10 == 0, f"Result ({result}) must be divisible by 10"
    return result
